/* k3p - show memory usage.
 * Public domain by Russell Marks, 2002.
 */

#include <PalmOS.h>
#include <Extensions/ExpansionMgr/VFSMgr.h>

#include "main-res.h"


#define K3P_CRID (('K'<<24)|('3'<<16)|('P'<<8)|'z')
#define K3P_PREFS_VER	1


#define BAR_WIDTH	158
#define MAX_BARS	4


struct prefs_tag
  {
  int showheap;
  int showvfs;
  int reserved[20];
  }
prefs;

int featuresetvfs;

static Char prnbuf[256];


void calc_and_draw(void);


void save_vars(void)
{
PrefSetAppPreferences(K3P_CRID,0,K3P_PREFS_VER,&prefs,sizeof(prefs),true);
}


void restore_vars(void)
{
UInt16 siz;
Int16 ver;

siz=sizeof(prefs);
ver=PrefGetAppPreferences(K3P_CRID,0,&prefs,&siz,true);
if(ver!=K3P_PREFS_VER || siz!=sizeof(prefs))
  {
  /* if we couldn't restore them, set up defaults */
  prefs.showheap=0;
  prefs.showvfs=1;
  MemSet(prefs.reserved,sizeof(prefs.reserved),0);
  }
}


Err start_app(void)
{
UInt32 romver,vfsver;

FtrGet(sysFtrCreator,sysFtrNumROMVersion,&romver);

/* must have OS >=2.0 */
if(romver<0x02003000)
  {
  FrmAlert(id_alert_romver);
  
  /* OS 1.0 will keep re-running us if we exit (!), so try the launcher. */
  SysAppLauncherDialog();
  return(1);
  }

featuresetvfs=0;
if(FtrGet(sysFileCVFSMgr,vfsFtrIDVersion,&vfsver)==0)
  featuresetvfs=1;	/* any version will do */

FrmGotoForm(id_main_form);

restore_vars();

return(0);
}


void stop_app(void)
{
save_vars();

FrmCloseAllForms();
}


#define get_obj_ptr(frm,id) \
	FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,id))


void do_prefs_form(void)
{
UInt16 button_id;
FormPtr frm;
int changed=0;

frm=FrmInitForm(id_prefs_form);

/* set up controls from prefs struct */
CtlSetValue(get_obj_ptr(frm,id_prefs_heap_check),prefs.showheap);
CtlSetValue(get_obj_ptr(frm,id_prefs_vfs_check),prefs.showvfs);

button_id=FrmDoDialog(frm);

/* copy from controls back to prefs struct */
if(button_id==id_prefs_ok_button)
  {
  int showheap,showvfs;
  
  showheap=CtlGetValue(get_obj_ptr(frm,id_prefs_heap_check));
  showvfs=CtlGetValue(get_obj_ptr(frm,id_prefs_vfs_check));
  if(showheap!=prefs.showheap || showvfs!=prefs.showvfs)
    changed=1;
  prefs.showheap=showheap;
  prefs.showvfs=showvfs;
  }

FrmDeleteForm(frm);

if(changed)
  calc_and_draw();
}


Boolean main_form_handle_event(EventType *event)
{
Boolean handled=false;

switch(event->eType)
  {
  case frmOpenEvent:
  case frmUpdateEvent:
    FrmDrawForm(FrmGetActiveForm());
    calc_and_draw();
    handled=true;
    break;
  
  case menuEvent:
    switch(event->data.menu.itemID)
      {
      case id_main_menu_options_rescan:
        calc_and_draw();
        handled=true;
        break;
        
      case id_main_menu_options_prefs:
        do_prefs_form();
        handled=true;
        break;

      case id_main_menu_options_about:
        FrmAlert(id_alert_options_about);
        handled=true;
        break;
      }
    break;
  
  default:
    break;
  }

return(handled);
}


Boolean app_handle_event(EventType *event)
{
FormPtr frm;
UInt16 form_id;
Boolean handled=false;

if(event->eType==frmLoadEvent)
  {
  /* init/activate form */
  form_id=event->data.frmLoad.formID;
  frm=FrmInitForm(form_id);
  FrmSetActiveForm(frm);

  /* set event handler for form */
  switch(form_id)
    {
    case id_main_form:
      FrmSetEventHandler(frm,main_form_handle_event);
      break;
    }

  handled=true;
  }

return(handled);
}


int showbar(int y,char *title,UInt32 avail,UInt32 total)
{
UnderlineModeType prevunderline=WinSetUnderlineMode(noUnderline);
FontID prevfont=FntSetFont(stdFont);
int fw,fh=FntCharHeight();
UInt32 availk,totalk,limit,pcnt;
RectangleType rect;
int ret=0;

if(!y && !title)
  ret=(fh+5)*2+2;
else
  {
  availk=avail/1024;		/* not rounded up */
  totalk=(total+1023)/1024;

  /* just in case, to avoid any divide-by-zero */
  if(!totalk) totalk++;
  
  /* we do the percentage maths etc. on the K totals to avoid
   * any risk of overflow (which otherwise only needs to go a bit past
   * 40MB free to cause trouble).
   */
  pcnt=(availk*100)/totalk;
  limit=((totalk-availk)*BAR_WIDTH)/totalk;

  /* if there's no title text, we assume something usable is
   * already in prnbuf.
   */
  if(title)
    StrPrintF(prnbuf,"%s",title);
  WinDrawChars(prnbuf,StrLen(prnbuf),0,y);

  StrPrintF(prnbuf,"%ldk/%ldk, %ld%% free",availk,totalk,pcnt);
  fw=FntCharsWidth(prnbuf,StrLen(prnbuf));
  WinDrawChars(prnbuf,StrLen(prnbuf),160-fw,y);

  y+=fh+5;
  rect.topLeft.x=0;
  rect.topLeft.y=y;
  rect.extent.x=BAR_WIDTH+2;
  rect.extent.y=fh;
  WinDrawRectangle(&rect,0);

  if(limit<BAR_WIDTH)
    {
    rect.topLeft.x=1+limit;
    rect.topLeft.y=y+1;
    rect.extent.x=BAR_WIDTH-limit;
    rect.extent.y=fh-2;
    WinEraseRectangle(&rect,0);
    }
  }

FntSetFont(prevfont);
WinSetUnderlineMode(prevunderline);
return(ret);
}


void calc_and_draw(void)
{
char *bar_text[MAX_BARS];
UInt32 bar_free[MAX_BARS],bar_size[MAX_BARS];
UInt16 bar_vfsid[MAX_BARS];
UInt16 nheaps,id,f,g;
UInt32 dsize,dfree,ssize,sfree;
UInt32 free_add,junk;
int nbars,barheight;
int top=22,left=160-top,ystart,yinc,y;
RectangleType rect;
UInt16 maxcards=MemNumCards();

dsize=dfree=ssize=sfree=0;

for(f=0;f<maxcards;f++)
  {
  nheaps=MemNumHeaps(f);
  for(g=0;g<nheaps;g++)
    {
    id=MemHeapID(f,g);
    /* ignore ROM */
    if(MemHeapFlags(id)&1)
      continue;
    MemHeapFreeBytes(id,&free_add,&junk);
    if(MemHeapDynamic(id))
      dsize+=MemHeapSize(id),dfree+=free_add;
    else
      ssize+=MemHeapSize(id),sfree+=free_add;
    }
  }

nbars=0;
bar_text[nbars]="Store";
bar_free[nbars]=sfree;
bar_size[nbars]=ssize;
nbars++;

if(featuresetvfs && prefs.showvfs)
  {
  UInt32 voli,vused,vsize;
  
  voli=vfsIteratorStart;
  while(voli!=vfsIteratorStop)
    {
    if(VFSVolumeEnumerate(&id,&voli)!=errNone)
      break;

    if(nbars>=MAX_BARS) break;

    if(VFSVolumeSize(id,&vused,&vsize)==errNone)
      {
      bar_text[nbars]=NULL;
      bar_vfsid[nbars]=id;
      bar_free[nbars]=vsize-vused;
      bar_size[nbars]=vsize;
      nbars++;
      }
    }
  }

if(prefs.showheap)
  {
  bar_text[nbars]="Heap";
  bar_free[nbars]=dfree;
  bar_size[nbars]=dsize;
  nbars++;
  }

rect.topLeft.x=0;
rect.topLeft.y=top;
rect.extent.x=160;
rect.extent.y=160-top;
WinEraseRectangle(&rect,0);

barheight=showbar(0,NULL,0,0);
yinc=barheight;
left-=nbars*yinc;
ystart=top;
if(left>0)
  {
  yinc+=left/nbars;
  ystart+=(left/nbars)/2;
  }

for(f=0,y=ystart;f<nbars;f++,y+=yinc)
  {
  if(bar_text[f])
    showbar(y,bar_text[f],bar_free[f],bar_size[f]);
  else
    {
    /* get VFS label */
    if(featuresetvfs &&
       VFSVolumeGetLabel(bar_vfsid[f],prnbuf,sizeof(prnbuf))==errNone)
      showbar(y,NULL,bar_free[f],bar_size[f]);
    else
      showbar(y,"VFS",bar_free[f],bar_size[f]);
    }
  }
}


void main_loop(void)
{
EventType event;
UInt16 error;

do
  {
  EvtGetEvent(&event,evtWaitForever);
  
  if(SysHandleEvent(&event))
    continue;
  
  if(MenuHandleEvent(NULL,&event,&error))
    continue;
  
  if(app_handle_event(&event))
    continue;

  FrmDispatchEvent(&event);
  }
while(event.eType!=appStopEvent);
}


UInt32 PilotMain(UInt16 cmd,MemPtr cmdPBP,UInt16 launchFlags)
{
Err ret=0;		
	
if(cmd==sysAppLaunchCmdNormalLaunch)
  {
  if((ret=start_app())==0)
    {
    main_loop();
    stop_app();
    }
  }

return(ret);
}
